import numpy as np

from fc_linalg3D.simplelib import is3Darray,getCpts,getMats,getVecs,getNdims,to3Darray

#
# ELEMENT BY ELEMENT OPERATIONS
#

# Generic element by element operations 
def genop_init(A,B,op):
  """  returns op(A,B) element by element operations
     computions if 
       - A and B are compatible 3D-arrays
       - A a 3D-array and B a scalar
       - B a 3D-array and A a scalar
  """
  isA3D=is3Darray(A)
  isB3D=is3Darray(B)
  assert isA3D or isB3D, 'One of the parameters must be a 3D-array'
  if isA3D and isB3D:
    assert A.shape==B.shape
    return op(A,B)
  if np.isscalar(A):
    return op(B,A)
  if np.isscalar(B):
    return op(A,B)
  return None

def check_dim(N,dimA,dimB):
  if dimA==dimB: # Both 3D-array
    return True
  NA,mA,nA=dimA; NB,mB,nB=dimB
  if (NA == N) and (NB == N): # one is a 3D-array, the other a (N,1,1) array
    return ( mB==1 and nB==1) or ( mA==1 and nA==1)
  if ( NA==1 ): # necessarly B is a 3D-array
    return ( mA==1 and nA==1 ) or (mA==mB and nA==nB)
  if ( NB==1 ): # necessarly A is a 3D-array
    return ( mB==1 and nB==1 ) or (mA==mB and nA==nB)
  return False

def genop_Cpt(A,B,op):
  """  returns op(A,B) element by element operation
  """
  #C=genop_init(A,B,op)
  #if C is not None:
    #return C
  getA,getB,N,dimA,dimB=getCpts(A,B)
  assert check_dim(N,dimA,dimB) 
  m=max(dimA[1],dimB[1]); n=max(dimA[2],dimB[2])
  C=np.zeros((N,m,n))
  for k in np.arange(N):
    for i in np.arange(m):
      for j in np.arange(n):
        C[k,i,j]=op(getA(A,k,i,j),getB(B,k,i,j))
  return C

def genop_Mat(A,B,op):
  """  returns op(A,B) element by element operation
  """
  #C=genop_init(A,B,op)
  #if C is not None:
    #return C
  getA,getB,N,dimA,dimB=getMats(A,B)
  assert check_dim(N,dimA,dimB) 
  m=max(dimA[1],dimB[1]); n=max(dimA[2],dimB[2])
  C=np.zeros((N,m,n))
  for k in np.arange(N):
    C[k]=op(getA(A,k),getB(B,k))
  return C

def genop_Vec(A,B,op):
  """  returns op(A,B) element by element operation
  """
  #C=genop_init(A,B,op)
  #if C is not None:
    #return C
  getA,getB,N,dimA,dimB=getVecs(A,B)
  assert check_dim(N,dimA,dimB) 
  m=max(dimA[1],dimB[1]); n=max(dimA[2],dimB[2])
  C=np.zeros((N,m,n))
  for i in np.arange(m):
    for j in np.arange(n):
      C[:,i,j]=op(getA(A,i,j),getB(B,i,j))
  return C

def genop_Cvt(A,B,op):
  """  returns op(A,B) element by element operation
  """
  #C=genop_init(A,B,op)
  #if C is not None:
    #return C
  N,dimA,dimB=getNdims(A,B)
  assert check_dim(N,dimA,dimB) 
  m=max(dimA[1],dimB[1]); n=max(dimA[2],dimB[2])
  A3D=to3Darray(A,N,m,n)
  B3D=to3Darray(B,N,m,n)
  return op(A3D,B3D)

def genop_Nat(A,B,op):
  """  returns op(A,B) element by element operation
  """
  isA3D=is3Darray(A)
  isB3D=is3Darray(B)
  assert isA3D or isB3D, 'One of the parameters must be a 3D-array'
  if isA3D and isB3D:
    assert A.shape==B.shape
    return op(A,B)
  if np.isscalar(A) or np.isscalar(B):
    return op(A,B)
  if isA3D and A.shape[0]==B.shape[0]:
    if len(B.shape)==2:
      return op(A,B.reshape([B.shape[0],B.shape[1],1]))
    else:
      return op(A,B.reshape([B.shape[0],1,1]))
  if isB3D and A.shape[0]==B.shape[0]:
    if len(A.shape)==2:
      return op(A.reshape([A.shape[0],A.shape[1],1]),B)
    else:
      return op(A.reshape([A.shape[0],1,1]),B)
  if isA3D and len(B.shape)==2:
    return op(A,B.reshape([1,B.shape[0],B.shape[1]]))
  if isB3D and len(A.shape)==2:
    return op(A.reshape([1,A.shape[0],A.shape[1]]),B)
  assert False
 
  
# 
plus_Cpt=add_Cpt=lambda A,B: genop_Cpt(A,B,np.add)
plus_Mat=add_Mat=lambda A,B: genop_Mat(A,B,np.add)
plus_Vec=add_Vec=lambda A,B: genop_Vec(A,B,np.add)
plus_Cvt=add_Cvt=lambda A,B: genop_Cvt(A,B,np.add)
plus_Nat=add_Nat=lambda A,B: genop_Nat(A,B,np.add)
plus=add=add_Nat

minus_Cpt=sub_Cpt=lambda A,B: genop_Cpt(A,B,np.subtract)
minus_Mat=sub_Mat=lambda A,B: genop_Mat(A,B,np.subtract)
minus_Vec=sub_Vec=lambda A,B: genop_Vec(A,B,np.subtract)
minus_Cvt=sub_Cvt=lambda A,B: genop_Cvt(A,B,np.subtract)
minus_Nat=sub_Nat=lambda A,B: genop_Nat(A,B,np.subtract)
minus=sub=sub_Nat

times_Cpt=mul_Cpt=lambda A,B: genop_Cpt(A,B,np.multiply)
times_Mat=mul_Mat=lambda A,B: genop_Mat(A,B,np.multiply)
times_Vec=mul_Vec=lambda A,B: genop_Vec(A,B,np.multiply)
times_Cvt=mul_Cvt=lambda A,B: genop_Cvt(A,B,np.multiply)
times_Nat=mul_Nat=lambda A,B: genop_Nat(A,B,np.multiply)
times=times_Nat

rdivide_Cpt=truediv_Cpt=lambda A,B: genop_Cpt(A,B,np.divide)
rdivide_Mat=truediv_Mat=lambda A,B: genop_Mat(A,B,np.divide)
rdivide_Vec=truediv_Vec=lambda A,B: genop_Vec(A,B,np.divide)
rdivide_Cvt=truediv_Cvt=lambda A,B: genop_Cvt(A,B,np.divide)
rdivide_Nat=truediv_Nat=lambda A,B: genop_Nat(A,B,np.divide)
rdivide=rdivide_Nat

ldivide_Cpt=lambda A,B: genop_Cpt(B,A,np.divide)
ldivide_Mat=lambda A,B: genop_Mat(B,A,np.divide)
ldivide_Vec=lambda A,B: genop_Vec(B,A,np.divide)
ldivide_Cvt=lambda A,B: genop_Cvt(B,A,np.divide)
ldivide_Nat=lambda A,B: genop_Nat(B,A,np.divide)
ldivide=ldivide_Nat

def aXpbY_init(a,X,b,Y):
  isX3D=is3Darray(X)
  isY3D=is3Darray(Y)
  assert isX3D or isY3D, 'One of the parameters must be an 3D-array'
  assert np.isscalar(a) and np.isscalar(b)
  if isX3D and isY3D:
    assert X.shape==Y.shape

def aXpbY_Cpt(a,X,b,Y):
  """  returns a*X+b*Y
  """
  assert np.isscalar(a) and np.isscalar(b)
  getX,getY,N,dimX,dimY=getCpts(X,Y)
  assert check_dim(N,dimX,dimY) 
  m=max(dimX[1],dimY[1]); n=max(dimX[2],dimY[2])
  C=np.zeros((N,m,n))
  for k in range(N):
    for i in np.arange(m):
      for j in np.arange(n):
        C[k,i,j]=a*getX(X,k,i,j) + b*getY(Y,k,i,j)
  return C 

def aXpbY_Mat(a,X,b,Y):
  """  returns a*X+b*Y
  """
  assert np.isscalar(a) and np.isscalar(b)
  getX,getY,N,dimX,dimY=getMats(X,Y)
  assert check_dim(N,dimX,dimY) 
  m=max(dimX[1],dimY[1]); n=max(dimX[2],dimY[2])
  C=np.zeros((N,m,n))
  for k in range(N):
    C[k]=a*getX(X,k) + b*getY(Y,k)
  return C 

def aXpbY_Vec(a,X,b,Y):
  """  returns a*X+b*Y
  """
  assert np.isscalar(a) and np.isscalar(b)
  getX,getY,N,dimX,dimY=getVecs(X,Y)
  assert check_dim(N,dimX,dimY) 
  m=max(dimX[1],dimY[1]); n=max(dimX[2],dimY[2])
  C=np.zeros((N,m,n))
  for i in np.arange(m):
    for j in np.arange(n):
      C[:,i,j]=a*getX(X,i,j) + b*getY(Y,i,j)
  return C  

def aXpbY_Cvt(a,X,b,Y):
  """  returns a*X+b*Y
  """
  assert np.isscalar(a) and np.isscalar(b)
  N,dimX,dimY=getNdims(X,Y)
  assert check_dim(N,dimX,dimY) 
  m=max(dimX[1],dimY[1]); n=max(dimX[2],dimY[2])
  X3D=to3Darray(X,N,m,n)
  Y3D=to3Darray(Y,N,m,n)
  return a*X3D+b*Y3D

def aXpbY_Nat(a,X,b,Y):
  """  returns a*X+b*Y
  """
  assert np.isscalar(a) and np.isscalar(b)
  isX3D=is3Darray(X)
  isY3D=is3Darray(Y)
  assert isX3D or isY3D, 'One of the parameters must be a 3D-array'
  if isX3D and isY3D:
    assert X.shape==Y.shape
    return a*X+b*Y
  if np.isscalar(X) or np.isscalar(Y):
    return a*X+b*Y
  if isX3D and X.shape[0]==Y.shape[0]:
    if len(Y.shape)==2:
      Ytmp=Y.reshape([Y.shape[0],Y.shape[1],1])
    else:
      Ytmp=Y.reshape([Y.shape[0],1,1])
    return a*X+b*Ytmp
  if isY3D and X.shape[0]==Y.shape[0]:
    if len(X.shape)==2:
      Xtmp=X.reshape([X.shape[0],X.shape[1],1])
    else:
      Xtmp=X.reshape([X.shape[0],1,1])
    return a*Xtmp+b*Y
  if isX3D and len(Y.shape)==2:
    Ytmp=Y.reshape([1,Y.shape[0],Y.shape[1]])
    return a*X+b*Ytmp
  if isY3D and len(X.shape)==2:
    Xtmp=X.reshape([1,X.shape[0],X.shape[1]])
    return a*Xtmp+b*Y
  assert False


def aXpbY(a,X,b,Y):
  return aXpbY_Nat(a,X,b,Y)
