import numpy as np

#
# CHOLESKY FACTORIZATION
#
def cholesky_std(A):
  """ Cholesky factorization A = B @ B^*
  """
  (m,n)=A.shape
  assert m==n
  B=np.zeros((n,n))
  for j in range(n):
    S=0
    for l in range(j):
      S+=np.abs(B[j,l]*B[j,l])
    B[j,j]=np.sqrt(A[j,j]-S)
    for i in range(j+1,n):
      S=0
      for l in range(j):
        S+=B[i,l]*np.conj(B[j,l])
      B[i,j]=(A[i,j]-S)/B[j,j]
  return B

def cholesky_std01(A):
  """ Cholesky factorization A = B @ B^*
  """
  (m,n)=A.shape
  assert m==n
  B=np.zeros((n,n))
  for j in range(n):
    B[j,j]=np.sqrt(A[j,j]-np.sum(B[j,:j]**2))
    for i in range(j+1,n):
      B[i,j]=(A[i,j]-np.sum(B[i,:j]*B[j,:j].conj()))/B[j,j]
  return B

def cholesky_Mat(A):
  """ Cholesky factorizations on (N,d,d) arrays such that
        A[k] = B[k] @ B[k]^*
  """
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  L=np.zeros((N,m,m))
  for k in np.arange(N):
    L[k]=cholesky_std(A[k])
  return L

def cholesky_Mat01(A):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  L=np.zeros((N,m,m))
  for k in np.arange(N):
    L[k]=cholesky_std01(A[k])
  return L

def cholesky_Cpt(A):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  L=np.zeros((N,m,m))
  for k in np.arange(N):
    for j in range(n):
      S=0
      for l in range(j):
        S+=np.abs(L[k,j,l]*L[k,j,l])
      L[k,j,j]=np.sqrt(A[k,j,j]-S)
      for i in range(j+1,n):
        S=0
        for l in range(j):
          S+=L[k,i,l]*np.conj(L[k,j,l])
        L[k,i,j]=(A[k,i,j]-S)/L[k,j,j]
  return L

def cholesky_Vec(A):
  """ Cholesky factorization A = B B^*
  """
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  B=np.zeros((N,n,n))
  S=np.zeros(N)
  for j in range(n):
    S*=0
    for k in range(j):
      S+=abs(B[:,j,k])**2
    B[:,j,j]=np.sqrt(A[:,j,j]-S)
    for i in range(j+1,n):
      S*=0
      for k in range(j):
        S+=B[:,i,k]*np.conj(B[:,j,k])
      B[:,i,j]=(A[:,i,j]-S)/B[:,j,j]
  return B

def cholesky_Vec01(A):
  """ Cholesky factorization A = B B^*
  """
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  B=np.zeros((N,n,n))
  S=np.zeros(N)
  for j in range(n):
    B[:,j,j]=np.sqrt(A[:,j,j]-np.sum(abs(B[:,j,:j])**2,axis=1))
    for i in range(j+1,n):
      B[:,i,j]=(A[:,i,j]-np.sum(B[:,i,:j]*B[:,j,:j].conj(),axis=1))/B[:,j,j]
  return B

def cholesky_Nat(A):
  return np.linalg.cholesky(A)

def pALU(A,inplace=False):
  (m,n)=A.shape
  assert m==n
  if not inplace:
    Ap=A.copy()
  else:
    Ap=A
  p=np.arange(n)
  for k in np.arange(n-1):
    mu=np.argmax(abs(Ap[k:,k]))+k
    if mu!=k:
      t=p[k];p[k]=p[mu];p[mu]=t
      Ap[[k,mu]]=Ap[[mu,k]]
    #Ap[k+1:,k]=Ap[k+1:,k]/Ap[k,k]
    for j in np.arange(k+1,m):
      Ap[j,k]=Ap[j,k]/Ap[k,k]
      Ap[j,k:]+=-Ap[j,k]*Ap[k,k:] # * not matricial operation
  if inplace:  
    return p,Ap
  else:
    L=np.tril(Ap,-1)
    L[np.arange(n),np.arange(n)]=1
    U=np.triu(Ap)
    return p,L,U
  
def LU(A):
  (m,n)=A.shape
  assert m==n
  U=np.zeros((m,m))
  L=np.eye(m)
  for i in np.arange(m):
    for j in np.arange(i,m):
      U[i,j]=A[i,j]-np.dot(L[i,:i],U[:i,j])
    for j in np.arange(i+1,m):
      L[j,i]=(A[j,i]-np.dot(L[j,:i],U[:i,i]))/U[i,i]
  return L,U

def pALU_10(A): # Bidouille!!!
  (m,n)=A.shape
  assert m==n
  U=np.zeros((m,m))
  L=np.eye(m)
  p=np.arange(n)
  for i in np.arange(m):
    S=np.dot(L[i,:i],U[:i,i])
    mu=np.argmax(abs(A[p[i:],i]-S))+i
    if mu!=i:
      t=p[i];p[i]=p[mu];p[mu]=t
      L[[i,mu],:i]=L[[mu,i],:i]
      S=np.dot(L[i,:i],U[:i,i])
    U[i,i]=A[p[i],i]-S
    for j in np.arange(i+1,m):
      U[i,j]=A[p[i],j]-np.dot(L[i,:i],U[:i,j])
      L[j,i]=(A[p[j],i]-np.dot(L[j,:i],U[:i,i]))/U[i,i]
  return p,L,U
  
def PALU_std(A):
  """ 
  """
  (m,n)=A.shape
  assert m==n
  U=A.copy()
  P=np.eye(m)
  L=np.eye(m)
  for k in np.arange(m-1):
    mu=np.argmax(abs(U[k:,k]))+k
    if abs(U[mu,k])> 1e-15:
      if mu!=k:
        U[[k,mu],k:]=U[[mu,k],k:]
        L[[k,mu],:k]=L[[mu,k],:k]
        P[[k,mu]]=P[[mu,k]]
      for j in np.arange(k+1,m):
        L[j,k]=U[j,k]/U[k,k]
        U[j,k:]+=-L[j,k]*U[k,k:]
  return P,L,U

def pALU_01(A):
  (m,n)=A.shape
  assert m==n
  U=A.copy()
  p=np.arange(n)
  L=np.eye(m)
  for k in np.arange(m-1):
    mu=np.argmax(abs(U[k:,k]))+k
    if mu!=k:
      U[[k,mu],k:]=U[[mu,k],k:]
      L[[k,mu],:k]=L[[mu,k],:k]
      t=p[k];p[k]=p[mu];p[mu]=t
    for j in np.arange(k+1,m):
      L[j,k]=U[j,k]/U[k,k]
      U[j,k:]+=-L[j,k]*U[k,k:]
  return p,L,U

def pALU_02(A):
  (m,n)=A.shape
  assert m==n
  U=A.copy()
  p=np.arange(n)
  L=np.eye(m)
  for k in np.arange(m-1):
    mu=np.argmax(abs(U[k:,k]))+k
    if abs(U[mu,k])> 1e-15:
      if mu!=k:
        U[[k,mu],k:]=U[[mu,k],k:]
        L[[k,mu],:k]=L[[mu,k],:k]
        t=p[k];p[k]=p[mu];p[mu]=t
      L[k+1:,k]=U[k+1:,k]/U[k,k]
      for j in np.arange(k+1,m):
        U[j,k:]+=-L[j,k]*U[k,k:]
  return p,L,U

def pALU_v00(A):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  U=zeros(N,m,m)
  p=np.zeros((N,m),dtype=np.int)
  L=eye(N,m)
  
  for k in np.arange(N):
    p[k],L[k],U[k]=pALU_02(A[k])
  return p,L,U

def PALU_Cpt(A):
  from fc_linalg3D.simplelib import is3Darray,eye
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  U=A.copy();P=eye(N,m);L=eye(N,m)
  for k in np.arange(N):
    for i in np.arange(m-1):
      mu=np.argmax(abs(U[k,i:,i]))+i
      if abs(U[k,mu,i])> 1e-15:
        if mu!=i:
          U[k,[i,mu],i:]=U[k,[mu,i],i:]
          L[k,[i,mu],:i]=L[k,[mu,i],:i]
          P[k,[i,mu]]=P[k,[mu,i]]
        for j in np.arange(i+1,m):
          L[k,j,i]=U[k,j,i]/U[k,i,i]
          U[k,j,i:]+=-L[k,j,i]*U[k,i,i:]
  return P,L,U

def PALU_Mat(A):
  from fc_linalg3D.simplelib import is3Darray,zeros
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  U=zeros(N,m,m);P=zeros(N,m,m);L=zeros(N,m,m)
  for k in np.arange(N):
    P[k],L[k],U[k]=PALU_std(A[k])
  return P,L,U

def PALU_Vec(A):
  from fc_linalg3D.simplelib import is3Darray,eye
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  U=A.copy();P=eye(N,m);L=eye(N,m)
  K=np.arange(N)
  J=np.ones((N,),dtype=np.int)
  for k in np.arange(m-1):
    mu=np.argmax(abs(U[:,k:,k]),1)+k
    U[[K,K],[k*J,mu],k:]=U[[K,K],[mu,k*J],k:]
    L[[K,K],[k*J,mu],:k]=L[[K,K],[mu,k*J],:k]
    P[[K,K],[k*J,mu]]=P[[K,K],[mu,k*J]]
    idx=np.where(abs(U[:,k,k])>1e-15)
    for i in np.arange(k+1,m): 
      L[idx,i,k]=U[idx,i,k]/U[idx,k,k]
      for j in np.arange(k,m):
        U[idx,i,j]+=-L[idx,i,k]*U[idx,k,j]
  return P,L,U


def pALU_v01(A):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  U=A.copy()
  p=np.tile(np.arange(n),(N,1))
  L=eye(N,m)
  K=np.arange(N)
  J=np.ones((N,),dtype=np.int)
  for k in np.arange(m-1):
    mu=np.argmax(abs(U[:,k:,k]),1)+k
    U[[K,K],[k*J,mu],k:]=U[[K,K],[mu,k*J],k:]
    L[[K,K],[k*J,mu],:k]=L[[K,K],[mu,k*J],:k]
    p[[K,K],[k*J,mu]]=p[[K,K],[mu,k*J]]
    idx=np.where(abs(U[:,k,k])>1e-15)
    for i in np.arange(k+1,m): 
      L[idx,i,k]=U[idx,i,k]/U[idx,k,k]
      for j in np.arange(k,m):
        U[idx,i,j]+=-L[idx,i,k]*U[idx,k,j]
  return p,L,U

def pLUinplace_std(A):
  (m,n)=A.shape
  assert m==n
  p=np.arange(m,dtype=np.int)
  for i in np.arange(m-1):
    mu=np.argmax(abs(A[i:,i]))+i
    if abs(A[mu,i])> 1e-15:
      if mu!=i:
        A[[i,mu]]=A[[mu,i]]
        p[[i,mu]]=p[[mu,i]]
      A[i+1:,i]=A[i+1:,i]/A[i,i]
      A[i+1:,i+1:]+=-np.dot(A[i+1:,i:i+1],A[i:i+1,i+1:]) # A[i+1:,i].shape ==(r,)  but A[i+1:,i:i+1].shape ==(r,1) is OK for dot
  return p,A

def pLUinplace_Mat(A):
  from fc_linalg3D.simplelib import is3Darray,zeros
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  p=np.zeros((N,m),dtype=np.int)
  for k in np.arange(N):
    p[k],A[k]=pLUinplace_std(A[k])
  return p,A

def pLUinplace_Cpt(A):
  from fc_linalg3D.simplelib import is3Darray,zeros
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  p=np.tile(np.arange(m,dtype=np.int),[N,1])
  for k in np.arange(N):
    for i in np.arange(m-1):
      mu=np.argmax(abs(A[k,i:,i]))+i
      if abs(A[k,mu,i])> 1e-15:
        if mu!=i:
          A[k,[i,mu]]=A[k,[mu,i]]
          p[k,[i,mu]]=p[k,[mu,i]]
        A[k,i+1:,i]=A[k,i+1:,i]/A[k,i,i]
        A[k,i+1:,i+1:]+=-np.dot(A[k,i+1:,i:i+1],A[k,i:i+1,i+1:])
  return p,A

def pLUinplace_Vec(A):
  from fc_linalg3D.simplelib import is3Darray,zeros
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  p=np.tile(np.arange(m,dtype=np.int),[N,1])
  K=np.arange(N,dtype=np.int)
  J=np.ones((N,),dtype=np.int)
  
  for i in np.arange(m-1):
    mu=np.argmax(abs(A[:,i:,i]),1)+i
    A[[K,K],[i*J,mu]]=A[[K,K],[mu,i*J]]
    p[[K,K],[i*J,mu]]=p[[K,K],[mu,i*J]]
    idx=np.where(abs(A[:,i,i])>1e-15)[0]
    I=np.arange(i+1,m)
    for j in np.arange(i+1,m):
      A[idx,j,i]=A[idx,j,i]/A[idx,i,i]
      for l in np.arange(i+1,m):
        A[idx,j,l]+= - A[idx,j,i]*A[idx,i,l]
  return p,A

def pLUinplace_Vec01(A):
  from fc_linalg3D.simplelib import is3Darray,zeros
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  p=np.tile(np.arange(m,dtype=np.int),[N,1])
  K=np.arange(N,dtype=np.int)
  J=np.ones((N,),dtype=np.int)
  
  for i in np.arange(m-1):
    mu=np.argmax(abs(A[:,i:,i]),1)+i
    A[[K,K],[i*J,mu]]=A[[K,K],[mu,i*J]]
    p[[K,K],[i*J,mu]]=p[[K,K],[mu,i*J]]
    idx=np.where(abs(A[:,i,i])>1e-15)[0]
    A[idx,i+1:,i]=A[idx,i+1:,i]/A[idx,i,i:i+1]
    A[idx,i+1:,i+1:]+= - A[idx,i+1:,i:i+1]*A[idx,i:i+1,i+1:]
  return p,A

def pLUinplace_pVec01(A):
  (m,n,N)=A.shape
  assert m==n
  p=np.tile(np.arange(m,dtype=np.int),[N,1])
  K=np.arange(N,dtype=np.int)
  J=np.ones((N,),dtype=np.int)
  
  for i in np.arange(m-1):
    mu=np.argmax(abs(A[i:,i,:]),0)+i
    for j in np.arange(m):
      A[[i*J,mu],j,[K,K]]=A[[mu,i*J],j,[K,K]]
      p[[K,K],[i*J,mu]]=p[[K,K],[mu,i*J]]
    idx=np.where(abs(A[i,i,:])>1e-15)[0]
    A[i+1:,i:i+1,idx]=A[i+1:,i:i+1,idx]/A[i:i+1,i:i+1,idx]
    A[i+1:,i+1:,idx]+= - A[i+1:,i:i+1,idx]*A[i:i+1,i+1:,idx]
  return p,A


def LUfrominplace(A):
  from fc_linalg3D.simplelib import is3Darray,triu,tril,eye
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  U=triu(A)
  L=tril(A,-1)+eye(N,m)
  return L,U

def permInd2Mat(p):
  from fc_linalg3D.simplelib import zeros
  N,d=p.shape
  K=np.tile(np.arange(N,dtype=np.int).reshape((N,1)),[1,d])
  I=np.tile(np.arange(d,dtype=np.int),[N,1])
  P=zeros(N,d,d)
  P[K,I,p]=1
  return P

def permsign(p):
  from fc_tools.math import combs
  (N,d)=p.shape
  I=combs(np.arange(d),2)
  n=I.shape[0]
  S=np.zeros((N,))
  for i in np.arange(n):
    S+=np.float64(p[:,I[i,0]]>p[:,I[i,1]])
  #S=1-2*np.remainder(S,2)
  S=1-2*np.fmod(S,2)
  return S
