import numpy as np

TypeArray=['array3d','array2d','array1d','constant']

def is3Darray(A):
  if not isinstance(A,np.ndarray):
    return False
  if len(A.shape) != 3:
    return False
  (N,m,n)=A.shape
  return N>m and N>n

def to3Darray(A,N,m,n): # A is (N,) or (m,n) array
  if is3Darray(A):
    assert A.shape==(N,m,n), '3Darray size failed'
    return A
  if np.isscalar(A):
    return A*np.ones((N,m,n))
  if len(A.shape)==1:
    assert A.shape[0]==N
    return np.swapaxes(np.swapaxes(np.tile(A,(m,n,1)),0,1),0,2)
  if len(A.shape)==2:
    assert A.shape==(m,n)
    return np.tile(A,(N,1,1))
  raise ValueError('Unable to convert array (%d,%d) in a 3Darray (%d,%d,%d)'%A.shape[0],A.shape[1],N,m,n)

def genComp_ebye(A,Type,**kwargs):
  rand=kwargs.get('rand', np.random.randn )
  Type=Type.lower()
  assert is3Darray(A) and Type in ['array3d','array2d','array1d','constant']
  (N,m,n)=A.shape
  if Type=='array3d':
    return rand(N,m,n)
  if Type=='array2d':
    return rand(m,n)  
  if Type=='array1d':
    return rand(N)
  return rand()
  
def dim3D(A):
  if np.isscalar(A):
    return (1,1,1) # N,m,n
  if isinstance(A,np.ndarray):
    n=A.shape
    if len(n)==2:
      return (1,n[0],n[1])
    if len(n)==3:
      return (n[0],n[1],n[2])
  raise ValueError() 

def getCpt(A,N):
  if np.isscalar(A):
    fun=lambda X,k,i,j: X
    return (fun,(1,1,1))
  assert isinstance(A,np.ndarray)
  n=A.shape
  if len(n)==1:
    fun=lambda X,k,i,j: X[k]
    if n[0]==N:
      return (fun,(n[0],1,1))
    else:
      return (fun,(1,n[0],1))
  if len(n)==2:
    if (n[0]==N and n[1]==1) or (n[0]==1 and n[1]==N):
      fun=lambda X,k,i,j: X[k]
      return (fun,(N,1,1))
    fun=lambda X,k,i,j: X[i,j]
    return (fun,(1,n[0],n[1]))
  if len(n)==3:
    fun=lambda X,k,i,j: X[k,i,j]
    return (fun,n)
  raise ValueError()

def getCpts(A,B):
  isA3D=is3Darray(A)
  isB3D=is3Darray(B)
  assert isA3D or isB3D, 'One of the parameters must be an 3D-array'
  if isA3D:
    dimA=dim3D(A)
    N=dimA[0]
    funA=lambda X,k,i,j: X[k,i,j]
    funB,dimB=getCpt(B,N)
  else:
    dimB=dim3D(B)
    N=dimB[0]
    funB=lambda X,k,i,j: X[k,i,j]
    funA,dimA=getCpt(A,N)
  return (funA,funB,N,dimA,dimB)

def getMat(A,N):
  if np.isscalar(A):
    fun=lambda X,k: X
    return (fun,(1,1,1))
  assert isinstance(A,np.ndarray)
  n=A.shape
  if len(n)==1:
    fun=lambda X,k: X[k]
    if n[0]==N:
      return (fun,(n[0],1,1))
    else:
      return (fun,(1,n[0],1))
  if len(n)==2:
    if (n[0]==N and n[1]==1) or (n[0]==1 and n[1]==N):
      fun=lambda X,k: X[k]
      return (fun,(N,1,1))
    fun=lambda X,k: X
    return (fun,(1,n[0],n[1]))
  if len(n)==3:
    fun=lambda X,k: X[k]
    return (fun,n)
  raise ValueError()

def getScaComp(A,N): # for compatibility with old release
  return getMat(A,N)

def getVec(A,N):
  if np.isscalar(A):
    fun=lambda X,i,j: X
    return (fun,(1,1,1))
  assert isinstance(A,np.ndarray)
  n=A.shape
  if len(n)==1:
    fun=lambda X,i,j: X
    if n[0]==N:
      return (fun,(n[0],1,1))
    else:
      return (fun,(1,n[0],1))
  if len(n)==2:
    if (n[0]==N and n[1]==1) or (n[0]==1 and n[1]==N):
      fun=lambda X,i,j: X
      return (fun,(N,1,1))
    fun=lambda X,i,j: X[i,j]
    return (fun,(1,n[0],n[1]))
  if len(n)==3:
    fun=lambda X,i,j: X[:,i,j]
    return (fun,n)
  raise ValueError()

def getComp(A,N): # for compatibility with old release
  return getVec(A,N)

def getMats(A,B):
  isA3D=is3Darray(A)
  isB3D=is3Darray(B)
  assert isA3D or isB3D, 'One of the parameters must be an 3D-array'
  if isA3D:
    dimA=dim3D(A)
    N=dimA[0]
    funA=lambda X,k: X[k]
    funB,dimB=getMat(B,N)
  else:
    dimB=dim3D(B)
    N=dimB[0]
    funB=lambda X,k: X[k]
    funA,dimA=getMat(A,N)
  return (funA,funB,N,dimA,dimB)

def getScaComps(A,B): # for compatibility with old release
  return getMats(A,B)

def getVecs(A,B):
  isA3D=is3Darray(A)
  isB3D=is3Darray(B)
  assert isA3D or isB3D, 'One of the parameters must be an 3D-array'
  if isA3D:
    N=A.shape[0]
  else:
    N=B.shape[0]
  funA,dimA=getVec(A,N)
  funB,dimB=getVec(B,N)
  return (funA,funB,N,dimA,dimB)

def getComps(A,B): # for compatibility with old release
  return getVecs(A,B)

def getDimension(A,N):
  if np.isscalar(A):
    return (1,1,1)
  assert isinstance(A,np.ndarray)
  n=A.shape
  if len(n)==1:
    if n[0]==N:
      return (n[0],1,1)
    else:
      return (1,n[0],1)
  if len(n)==2:
    if (n[0]==N and n[1]==1) or (n[0]==1 and n[1]==N):
      return (N,1,1)
    return (1,n[0],n[1])
  if len(n)==3:
    assert( n[0]==N )
    return n
  raise ValueError()

def getNdims(A,B):
  isA3D=is3Darray(A)
  isB3D=is3Darray(B)
  assert isA3D or isB3D, 'One of the parameters must be an 3D-array'
  if isA3D:
    N=A.shape[0]
  else:
    N=B.shape[0]
  dimA=getDimension(A,N)
  dimB=getDimension(B,N)
  return (N,dimA,dimB)

def transpose(A):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  return np.transpose(A,(0,2,1))

def ctranspose(A):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  return np.conj(np.transpose(A,(0,2,1)))

def trace(A):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  T=np.zeros(N)
  for i in np.range(m):
    T+=A[:,i,i]

def diag(A,k=0):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  D=np.zeros((N,m,n))
  for i in np.arange(min(m,n)):
    j=i+k
    if j>=0 and j<n:
      D[:,i,j]=A[:,i,j]
  return D

def triu(A,k=0):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  D=np.zeros((N,m,n))
  for i in np.arange(m):
    D[:,i,max(0,min(i+k,n)):]=A[:,i,max(0,min(i+k,n)):]
  return D

def tril(A,k=0):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  D=np.zeros((N,m,n))
  for i in np.arange(m):
    D[:,i,:min(n,max(0,i+k+1))]=A[:,i,:min(n,max(0,i+k+1))]
  return D
 
def eye(N,d,**kwargs):
  A=np.zeros((N,d,d),**kwargs)
  I=np.ones((N,),**kwargs)
  for i in np.arange(d):
    A[:,i,i]=I
  return A

def ones(N,m,n,**kwargs):
  return np.ones((N,m,n),**kwargs)

def zeros(N,m,n,**kwargs):
  return np.zeros((N,m,n),**kwargs)

def diagprod(A):
  assert is3Darray(A), 'Parameter must be a 3D-array'
  (N,m,n)=A.shape
  assert m==n
  D=np.ones((N,))
  for i in np.arange(m):
    D=D*A[:,i,i]
  return D
